# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""C{agg [-g GROUPING_FUNCTION] INITIAL_VALUE AGGREGATOR}

C{agg [-c GROUPING_FUNCTION] INITIAL_VALUE AGGREGATOR}

Combines inputs into a smaller number of outputs. For example, if
the input is a stream of numbers, then C{agg} can be used to generate
output comprising the sum of the inputs.

If C{GROUPING_FUNCTION} is omitted, then one output object is
generated by initializing an accumulator to C{INITIAL_VALUE} and then
combining the accumulator with input objects using C{AGGREGATOR}.  The
inputs to C{AGGREGATOR} are the current values of the accumulator
followed by the members of the input object.

B{Example}: If the input objects are integers C{1, 2, 3}, then the sum of
the integers is computed by::

    agg 0 'sum, x: sum + x'

which yields C{6}.

B{Example}: The sum and sum of squares of a sequence of input number can be computed
by::

    agg '(0, 0)' 'sum, sum_squares, input: (sum + input, sum + input ** 2)'

In this case, the accumulator is a tuple containing two integers, the
sum and the sum of squares. The initial value of this accumulator is
C{(0, 0)}.  Each input consists of a single number. The aggregator
function takes as input the two parts of the accumulator followed by
the next number of the input, and outputs a tuple containing the
updated accumulator value.

If C{GROUPING_FUNCTION} is specified, then a set of accumulators is
maintained, one for each value of C{GROUPING_FUNCTION}. Each output
value consists of the group value followed by the accumulated values
for the group, ("values" not "value" because the accumulator may be a
sequence.)

Example: If the input objects are C{('a', 1), ('a', 2), ('b', 3), ('b', 4)}, then
the sum of ints for each string is computed by::

    agg -g 'label, number: label' 0 'sum, label, number: sum + number'

which yields C{('a', 3), ('b', 7)}.

If C{GROUPING_FUNCTION} is specified with the C{-g} flag, then C{agg}
generates its output when the input stream has ended. (It can't
generate output earlier, because group members map appear in any
order.) In some situations however, group members appear
consecutively, and it is useful to get output earlier. If group
members are known to be consecutive, then C{GROUPING_FUNCTION} can be
specified using the C{-c} flag.
"""

import osh.core
import osh.args

_wrap_if_necessary = osh.core.wrap_if_necessary
Option = osh.args.Option

# CLI
def _agg():
    return _Agg()

# API
def agg(initial_value, aggregator, group = None, consecutive = None):
    """Combine inputs into a smaller number of outputs. If neither C{group} nor
    C{consecutive} is specified, then there is one accumulator, initialized to
    C{initial_value}. The C{aggregator} function is used to combine the current value
    of the accumulator with the input to yield the next value of the accumulator.
    The arguments to C{aggregator} are the elements of the accumulator followed
    by the elements of one piece of input.
    If C{group} is specified, then there is one accumulator for each group value, defined
    by applying the function C{group} to each input. C{consecutive} is just like C{group}
    except that it is assumed that group values are adjacent in the input sequence.
    At most one of C{group} and C{consecutive} may be specified.
    """
    args = [initial_value, aggregator]
    if group:
        args.append(Option('-g', group))
    if consecutive:
        args.append(Option('-c', consecutive))
    return _Agg().process_args(*args)

class _Agg(osh.core.Op):

    _aggregate = None


    # object interface

    def __init__(self):
        osh.core.Op.__init__(self, 'g:c:', (2, 2))


    # BaseOp interface

    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        grouping_function = args.function_arg('-g')
        consecutive_grouping_function = args.function_arg('-c')
        initial_value = _wrap_if_necessary(args.next_eval())
        aggregation_function = args.next_function()
        if grouping_function and consecutive_grouping_function:
            self.usage()
        if initial_value is None or aggregation_function is None:
            self.usage()
        if args.has_next():
            self.usage()
        if grouping_function and consecutive_grouping_function:
            self.usage()
        elif grouping_function:
            self._aggregate = _GroupingAggregate(
                self,
                grouping_function,
                initial_value,
                aggregation_function)
        elif consecutive_grouping_function:
            self._aggregate = _ConsecutiveGroupingAggregate(
                self,
                consecutive_grouping_function,
                initial_value,
                aggregation_function)
        else:
            self._aggregate = _SimpleAggregate(
                self,
                initial_value,
                aggregation_function)


    def receive(self, object):
        self._aggregate.receive(object)

    def receive_complete(self):
        self._aggregate.receive_complete()


class _GroupingAggregate(object):
    _command = None
    _group_function = None
    _initial_value = None
    _aggregate_function = None
    _sum = None

    def __init__(self, command, group_function, initial_value, aggregate_function):
        self._command = command
        self._group_function = group_function
        self._initial_value = initial_value
        self._aggregate_function = aggregate_function
        self._sum = {}

    def receive(self, object):
        group = self._group_function(*object)
        sum = self._sum.get(group, self._initial_value)
        self._sum[group] = _wrap_if_necessary(self._aggregate_function(*(tuple(sum) + tuple(object))))

    def receive_complete(self):
        for group, sum in self._sum.iteritems():
            self._command.send(_wrap_if_necessary(group) + tuple(sum))
        self._command.send_complete()

class _ConsecutiveGroupingAggregate(object):
    _command = None
    _group_function = None
    _initial_value = None
    _aggregate_function = None
    _group = None
    _sum = None

    def __init__(self, command, group_function, initial_value, aggregate_function):
        self._command = command
        self._group_function = group_function
        self._initial_value = initial_value
        self._aggregate_function = aggregate_function
        self._group = None
        self._sum = None

    def receive(self, object):
        newGroup = self._group_function(*object)
        if self._group is None or self._group != newGroup:
            if self._group is not None:
                self._command.send(_wrap_if_necessary(self._group) + tuple(self._sum))
            self._group = newGroup
            self._sum = self._initial_value
        self._sum = _wrap_if_necessary(self._aggregate_function(*(tuple(self._sum) + tuple(object))))

    def receive_complete(self):
        if self._group is not None:
            self._command.send(_wrap_if_necessary(self._group) + tuple(self._sum))
        self._command.send_complete()

class _SimpleAggregate(object):
    _command = None
    _initial_value = None
    _aggregate_function = None
    _sum = None

    def __init__(self, command, initial_value, aggregate_function):
        self._command = command
        self._initial_value = initial_value
        self._aggregate_function = aggregate_function
        self._sum = initial_value

    def receive(self, object):
        self._sum = _wrap_if_necessary(self._aggregate_function(*(tuple(self._sum) + tuple(object))))

    def receive_complete(self):
        self._command.send(self._sum)
        self._command.send_complete()
