# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""C{copyfrom [-Crpx] -c CLUSTER FILES LOCAL_DIR}

Copies C{FILES} from each node of the specified C{CLUSTER} to
C{LOCAL_DIR}.  If C{CLUSTER} is omitted, then the default cluster is
used.  C{FILES} identifies files on nodes of the cluster. It must be
an absolute path, (i.e. must begin with C{'/'}) and may specify
multiple files, e.g. C{/foo/bar*/*.[0-9]}.  C{LOCAL_DIR} is a
directory on the node from which the command is being issued.

C{copyfrom} will normally populate a directory under C{LOCAL_DIR} for
each node in the cluster, creating the subdirectories if necessary.
The name of the subdirectory will be the same as the node's name,
specified in C{.oshrc}. But if C{-x} is specified, then subdirectories
will not be populated; files will be placed directly in
C{LOCAL_DIR}. The C{-x} option is supported only for single-node
clusters.
    
C{copyfrom} is implemented using C{scp}, and the following C{scp} flags are supported:
    - C{-C}: enable compression.
    - C{-r}: recursive copy
    - C{-p}: preserve modification times, access times, and modes.
"""

import os
import pickle
import threading

import osh.args
import osh.core
import osh.spawn

Spawn = osh.spawn.Spawn
Option = osh.args.Option

# CLI
def _copyfrom():
    return _CopyFrom()

# API
def copyfrom(cluster, files, local_dir,
             compress = False, recursive = False, preserve = False, no_subdirs = False):
    """Copies C{files} from each node of the specified C{cluster} to C{local_dir}.
    If C{no_subdirs} is False, then a subdirectory under C{local_dir} is created for
    each node of the cluster and the files from a node are copied to that node's subdirectory.
    If C{no_subdirs} is true, then C{cluster} must be a single-node cluster, no subdirectory
    is created, and files are copied directly into C{local_dir}. Compression is used
    for copying if C{compress} is True. Directories are copied recursively if C{recursive} is
    True. File attributes are preserved if C{preserve} is True.
    """
    args = [Option('-c', cluster)]
    scp_options = ''
    if compress:
        scp_options += 'C'
    if recursive:
        scp_options += 'r'
    if preserve:
        scp_options += 'p'
    if len(scp_options) > 0:
        scp_options = '-' + scp_options
    args.append(Option(scp_options))
    if no_subdirs:
        args.append(Option('-x'))
    return _CopyFrom().process_args(*args)

# scp options, also used by copyto

def scp(options):
    if options[0] != '-':
        options = '-' + options
    return osh.args.Option(options)

class _CopyFrom(osh.core.RemoteOp):

    # state

    _local_dir = None
    _file = None
    _scp_options = None
    _use_subdirs = None


    # object interface
    
    def __init__(self):
        osh.core.RemoteOp.__init__(self, 'Crpxc:', (2, 2))

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.core.RemoteOp.setup(self)
        args = self.args()
        self._use_subdirs = not args.flag('-x')
        if not self._use_subdirs and len(self.hosts()) > 1:
            self.usage()
        if args.has_next():
            self._file = args.next_string()
        if args.has_next():
            self._local_dir = args.next_string()
        if not self._file or not self._local_dir or args.has_next():
            self.usage()
        if not self._file.startswith('/'):
            self.usage()

    # RemoteOp interface

    def action(self, host):
        target_dir = self._local_dir
        if self._use_subdirs:
            target_dir += '/' + host.name
        try:
            os.makedirs(target_dir)
        except OSError:
            pass
        _copydown(self.user(),
                  host.address,
                  self._file,
                  target_dir,
                  self.scp_options())

def _copydown(user, host, file, local_dir, options = ''):
    scp_command = 'scp %s %s@%s:%s %s' % (options, user, host, file, local_dir)
    Spawn(scp_command, None, None, None).run()


