# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""C{copyto [-rpC] -c CLUSTER FILE ... REMOTE_DIR}

Copies C{FILE}s to C{REMOTE_DIR} on each node of the specified
C{CLUSTER}. If C{CLUSTER} is omitted, then the default cluster is
used.

C{REMOTE_DIR} must be an absolute path, (i.e. must begin with C{'/'}).

C{copyto} is implemented using C{scp}, and the following C{scp} flags
are supported:
    - C{-C}: enable compression.
    - C{-r}: recursive copy
    - C{-p}: preserve modification times, access times, and modes.
"""

import pickle
import threading

import osh.core
import osh.spawn

Spawn = osh.spawn.Spawn

# CLI
def _copyto():
    return _CopyTo()

# API
def copyto(cluster, files,
           compress = False, recursive = False, preserve = False):
    """Copies files to each node of the specified C{cluster}. The last elements
    of C{files} is the target directory on each node. The preceding elements are
    the local files to be copied. Compression is used  for copying if C{compress}
    is True. Directories are copied recursively if C{recursive} is
    True. File attributes are preserved if C{preserve} is True.
    """
    if len(args) > 0:
        cluster = args[0]
        args = (osh.args.Option('-c', cluster),) + args[1:]
    return _CopyTo().process_args(*args)

class _CopyTo(osh.core.RemoteOp):

    # state

    _remote_dir = None
    _files = None
    _scp_options = None


    # object interface
    
    def __init__(self):
        osh.core.RemoteOp.__init__(self, 'rpCc:', (2, None))


    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.core.RemoteOp.setup(self)
        args = self._args.remaining()
        if len(args) < 2:
            self.usage()
        self._remote_dir = args[-1]
        if not self._remote_dir.startswith('/'):
            self.usage()
        self._files = args[:-1]
        for option in osh.core.SCP_OPTIONS:
            self._check_option(option)


    # RemoteOp interface

    def action(self, host):
        _copyup(self._files,
                self.user(),
                host.address,
                self._remote_dir,
                self.scp_options())

def _copyup(files, user, host, remote_dir, options = ''):
    if isinstance(files, list) or isinstance(files, tuple):
        files = ' '.join(files)
    scp_command = 'scp %s %s %s@%s:%s' % (options, files, user, host, remote_dir)
    Spawn(scp_command, None, None, None).run()

