# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""C{install [-c CLUSTER] [-p PACKAGE] DIRECTORY}

C{install [-c CLUSTER] [-p PACKAGE] PYTHON_MODULE ...}

Installs Python modules to each node of C{CLUSTER}.  The modules are
identified as individual C{PYTHON_MODULE}s, (i.e. filenames ending in
C{.py}), or as a C{DIRECTORY} containing the modules of interest. In
the latter case, the contents of sub-directories will be copied also,
as sub-packages of C{PACKAGE}.

Modules are installed into the remote C{site-packages} directory, in
the subdirectory specified by C{PACKAGE}, (or in C{site-packages}
directly, if C{PACKAGE} is omitted).


For example, to install the local C{foo.bar} package on cluster
C{fred}::

    osh install -c fred -p foo.bar ~/myproject/foo/bar

To install a single module, C{xyz}, directly in C{site-packages}::

    osh install -c fred ~/myproject/util/xyz.py

If C{CLUSTER} is omitted, then installation is done locally. 

Not available through the API.
"""

import os
import os.path
import sys

import osh.core
import osh.spawn
import progtrack

Spawn = osh.spawn.Spawn
SpawnSSH = osh.spawn.SpawnSSH
collect_lines = osh.spawn.collect_lines

# CLI
def _install():
    return _Install()

def _scp(user, host, flags, source, target):
    err_lines = []
    scp_command = 'scp %s %s %s@%s:%s' % (flags, source, user, host, target)
    Spawn(scp_command,
          None,
          None,
          collect_lines(err_lines)).run()
    if len(err_lines) > 0:
        raise Exception(' '.join(err_lines))

def _ssh(user, host, command):
    out_lines = []
    SpawnSSH(user,
             host,
             command,
             None,
             collect_lines(out_lines),
             None).run()
    return out_lines
    
class _Install(osh.core.RemoteOp):

    _ui = None
    _package = None
    _directory = None
    _modules = None


    # object interface

    def __init__(self):
        osh.core.RemoteOp.__init__(self, 'c:p:', (1, 1))

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.core.RemoteOp.setup(self,
                                False # cluster_required = False
                                )
        self._package = self.args().string_arg('-p')
        sources = self.args().remaining()
        if len(sources) == 1 and os.path.isdir(sources[0]):
            self._directory = sources[0]
        else:
            for source in sources:
                if not source.endswith('.py'):
                    self.usage()
            self._modules = [x for x in sources]


    # Generator interface

    def execute(self):
        if self.cluster():
            self._ui = progtrack.ProgressTrackingUI('install')
            self._ui.add_column('host', 25)
            self._ui.add_column(['find', 'install dir'], 12);
            self._ui.add_column('installed', 12)
            for host in self.hosts():
                self._ui.add_row(host.name)
            self._ui.start()
            try:
                osh.core.RemoteOp.execute(self)
            finally:
                self._ui.stop()
        else:
            self.install_local()


    # RemoteOp interface

    def action(self, host):
        try:
            # Find install directory
            stage = 1
            version_command = ""
            output = _ssh(self.user(),
                          host.address,
                          "python -c 'import sys; print (sys.prefix,) + sys.version_info[0:2]'")
            package_dir = self.package_dir()
            install_dir = self.install_dir(eval(output[0])) + '/' + package_dir
            if package_dir != '':
                _ssh(self.user(),
                     host.address,
                     'mkdir -p %s' % install_dir)
                     
            self._ui.ok(host.name, stage)
            # Copy files
            stage = 2
            if self._directory:
                flags = '-rp'
                sources = '%s/*' % self._directory
            elif self._modules:
                flags = '-p'
                sources = ' '.join(self._modules)
            else:
                assert False
            _scp(self.user(), host.address, flags, sources, install_dir)
            self._ui.ok(host.name, stage)
        except:
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            message = str(exc_value).strip()
            self._ui.error(host.name, stage, message)


    # For use by this class
    
    def install_local(self):
        install_dir = self.install_dir((sys.prefix,) + sys.version_info[0:2])
        if self._directory:
            flags = '-Rp'
            sources = '%s/*' % self._directory
        elif self._modules:
            flags = '-p'
            sources = ' '.join(self._modules)
        else:
            assert False
        os.system('cp %s %s %s' % (flags, sources, install_dir))

    def install_dir(self, install_info):
        return '/%s/lib/python%s.%s/site-packages' % install_info

    def package_dir(self):
        if self._package:
            path = self._package.replace('.', '/')
        else:
            path = ''
        return path
