# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""C{installosh -c CLUSTER [OSH_CONFIG_FILE]}

Installs osh on C{CLUSTER}. On each node of the cluster, a C{.oshrc}
files is created in the home directory of the user specified for
C{CLUSTER}. This file is copied from C{OSH_CONFIG_FILE}, if
specified. Otherwise, C{~/.oshrc} is copied.

Not available through the API.
"""

import os
import sys
import threading

import osh.core
import osh.config
import osh.spawn
import progtrack


_DISTRIBUTION_FILENAME = 'osh-' + osh.config.OSH_VERSION + '.tar.gz'
_CONFIG_FILE = '.oshrc'
_PACKAGE_HOME = '/usr/share/osh'

Spawn = osh.spawn.Spawn
SpawnSSH = osh.spawn.SpawnSSH
collect_lines = osh.spawn.collect_lines

def _scp(user, host, source, target):
    err_lines = []
    scp_command = 'scp %s %s@%s:%s' % (source, user, host, target)
    Spawn(scp_command,
          None,
          None,
          collect_lines(err_lines)).run()
    if len(err_lines) > 0:
        raise Exception(' '.join(err_lines))

def _ssh(user, host, command):
    out_lines = []
    SpawnSSH(user,
             host,
             command,
             None,
             collect_lines(out_lines),
             None).run()
    return out_lines

# CLI
def _installosh():
    return _InstallOsh()

class _InstallOsh(osh.core.RemoteOp):

    _config_file = None
    _host_counter = None
    _ui = None

    # object interface

    def __init__(self):
        osh.core.RemoteOp.__init__(self, 'c:', (0, 1))
    

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.core.RemoteOp.setup(self)
        if self.args().has_next():
            self._config_file = self.args().next_string()
        else:
            self._config_file = '%s/%s' % (os.environ['HOME'], _CONFIG_FILE)
        self._host_counter = 0
        self._ui = progtrack.ProgressTrackingUI('installosh')
        self._ui.add_column('host', 25)
        self._ui.add_column(['create', 'osh dir'], 10)
        self._ui.add_column(['copy', 'package'], 10)
        self._ui.add_column(['copy', 'installer'], 10)
        self._ui.add_column(['run', 'installer'], 10)
        self._ui.add_column(['copy', 'conf file'], 10)
        for host in self.hosts():
            self._ui.add_row(host.name)

    # generator interface

    def execute(self):
        self._ui.start()
        try:
            osh.core.RemoteOp.execute(self)
        finally:
            self._ui.stop()


    # RemoteOp interface

    def action(self, host):
        stage = 0
        try:
            # Remove existing /usr/share/osh directory
            stage += 1
            _ssh(self.user(),
                 host.address,
                 '"rm -rf %s"' % _PACKAGE_HOME)
            # Create directory /usr/share/osh
            _ssh(self.user(),
                 host.address,
                 '"mkdir %s"' % _PACKAGE_HOME)
            self._ui.ok(host.name, stage)
            # Copy package to /usr/share/osh
            stage += 1
            _scp(self.user(),
                 host.address,
                 '%s/%s' % (_PACKAGE_HOME, _DISTRIBUTION_FILENAME),
                 _PACKAGE_HOME)
            self._ui.ok(host.name, stage)
            # Copy remoteinstall script to /usr/share/osh
            stage += 1
            _scp(self.user(),
                 host.address,
                 '%s/remoteinstall' % _PACKAGE_HOME,
                 _PACKAGE_HOME)
            self._ui.ok(host.name, stage)
            # Run remoteinstall script
            stage += 1
            _ssh(self.user(),
                 host.address,
                 '"%s/remoteinstall %s"' % (_PACKAGE_HOME, _PACKAGE_HOME))
            self._ui.ok(host.name, stage)
            # Copy config file to home directory unless one already exists
            stage += 1
            ls_oshrc = _ssh(self.user(),
                            host.address,
                            '"ls /%s/%s"' % (self.user(), _CONFIG_FILE))
            if len(ls_oshrc) == 0:
                _scp(self.user(),
                     host.address,
                     self._config_file,
                     '~')
            self._ui.ok(host.name, stage)
        except:
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            message = str(exc_value).strip()
            self._ui.error(host.name, stage, message)
