# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""For API usage only, (for CLI use C{osh @CLUSTER [ ... ]} syntax instead.)
"""

import osh.args
import osh.core
import osh.spawn
import osh.error

Spawn = osh.spawn.Spawn
ObjectInputProvider = osh.spawn.ObjectInputProvider
ObjectOutputConsumer = osh.spawn.ObjectOutputConsumer
LineOutputConsumer = osh.spawn.LineOutputConsumer
Option = osh.args.Option

_REMOTE_EXECUTABLE = 'remoteosh'

def _dump(stream, object):
    stream.dump(object)

def _consume_remote_stdout(consumer, host, object):
    if isinstance(object, osh.error.Error):
        exception = object.recreate_exception()
        osh.error.exception_handler(exception, object.command_description(), object.input(), host)
    else:
        consumer.send((host,) + tuple(object))

def _consume_remote_stderr(consumer, host, line):
    # UGLY HACK: remoteosh can occasionally return "[Errno 9] Bad file descriptor" on stderr.
    # I think this is because of io to a process stream whose process has completed.
    # I haven't had luck in tracking this down and fixing the problem for real, so
    # this is a grotesque workaround.
    if '[Errno 9] Bad file descriptor' not in line:
        osh.error.stderr_handler(line, consumer, None, host)

# CLI
def _remote():
    return _Remote()

# API
def remote(cluster, remote_command):
    """Execute a command on each node of C{cluster}. Output from each node is sent to
    the output stream, with the node name added as the first element of the output tuple.
    """
    import osh.apiparser
    op = _Remote()
    if isinstance(remote_command, osh.core.Op):
        remote_command = [remote_command]
    remote_pipeline = osh.apiparser._sequence_op(remote_command)
    op.set_pipeline(remote_pipeline)
    return op.process_args(Option('-c', cluster))

class _Remote(osh.core.RemoteOp):

    # state

    _pipeline = None


    # object interface
    
    def __init__(self):
        osh.core.RemoteOp.__init__(self, 'c:', (0, 0))        

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.core.RemoteOp.setup(self)


    # generator interface

    def execute(self):
        osh.core.RemoteOp.execute(self)
        self.send_complete()


    # RemoteOp interface

    def action(self, host):
        process = Spawn(
            self._remote_command(host.address, self.user(), host.db_profile),
            ObjectInputProvider(lambda stream, object: _dump(stream, object),
                                [osh.core.verbosity, self._pipeline]),
            ObjectOutputConsumer(lambda object: _consume_remote_stdout(self, host.name, object)),
            LineOutputConsumer(lambda line: _consume_remote_stderr(self, host.name, line)))
        process.run()
        if process.terminating_exception():
            self.set_terminating_exception(process.terminating_exception())


    # remote compile-time interface

    def set_pipeline(self, pipeline):
        self._pipeline = pipeline
        return self


    # for use by this class

    def _remote_command(self, host, user, db_profile):
        if db_profile:
            remote_command = '"%s %s"' % (_REMOTE_EXECUTABLE, db_profile)
        else:
            remote_command = _REMOTE_EXECUTABLE
        ssh_command = 'ssh %s -l %s %s' % (host, user, remote_command)
        return ssh_command
