# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Objects comprising an osh command's parse tree. Also
includes some classes used at command runtime.
"""

import getopt
import new
import os
import popen2
import sys
import threading
import types
import traceback

from config import config_value, oshrc_symbols
import cluster
import error
import args

USAGE_API = 'api'
USAGE_CLI = 'cli'
# Assume API. If parser is used then usage will be changed.
osh_usage = USAGE_API

# scp options supported by copyfrom and copyto
SCP_OPTIONS = 'rpC'

# 2.2 threads sometimes hang without this. The 2.2 default is 10, 2.3: 100
sys.setcheckinterval(100)

verbosity = None
default_db_profile = None

_id_counter = 0
_id_counter_lock = threading.Condition()

def _next_id():
    global _id_counter
    global _id_counter_lock
    _id_counter_lock.acquire()
    _id_counter += 1
    next_id = _id_counter
    _id_counter_lock.release()
    return next_id

def wrap_if_necessary(object):
    if not(isinstance(object, tuple) or isinstance(object, list)):
        object = (object,)
    return object

#--------------------------------------------------------------------------------

class OshObject(object):
    """Base class for all objects in a parse tree representing an osh
    command.
    """

    # state

    _id = None

    
    # object interface

    def __init__(self):
        self._id = _next_id()


    # OshObject interface

    def id(self):
        """Unique identifier within a single command.
        """
        return self._id

    def dump(self):
        """Renders this object and subtree as a string for debugging purposes.
        """
        assert False

#--------------------------------------------------------------------------------

class BaseOp(OshObject):
    """Base class for all osh commands, and for pipelines, (sequences of
    commands). Methods of this class implement the command execution and
    inter-command communication. The send* commands are used by subclasses to
    send output to downstream commands. The receive* commands are implemented
    by subclasses to receive and process input from upstream commands.
    """

    # state

    _parent = None
    _next_op = None
    _receiver = None


    # object interface
    
    def __init__(self):
        OshObject.__init__(self)

    def __str__(self):
        assert False


    # OshObject interface
    
    def dump(self):
        buffer = []
        op = self
        while op:
            buffer.append(str(op))
            op = op._next_op
        return ' ^ '.join(buffer)        


    # BaseOp runtime interface

    def usage(self):
        """Print usage message and exit.
        """
        sys.stderr.write(self.doc())
        sys.stderr.write('\n')
        sys.exit(-1)

    def doc():
        """Print command usage information.
        """
        assert False
    
    def setup(self):
        """Setup is executed just prior to command execution.
        """
        assert False

    def execute(self):
        """Execute the command.
        """
        assert False

    def send(self, object):
        """Called by a command class to send an object of command output to
        the next command.
        """
        try:
            if self._receiver:
                self._receiver.receive(wrap_if_necessary(object))
        except error.ExceptionHandlerException:
            raise
        except Exception, e:
            error.exception_handler(e, self._receiver, object)

    def send_complete(self):
        """Called by a command class to indicate that there will
        be no more output from the command.
        """
        try:
            if self._receiver:
                self._receiver.receive_complete()
        except error.ExceptionHandlerException:
            raise
        except Exception, e:
            error.exception_handler(e, self._receiver, object)

    def receive(self, object):
        """Implemented by a command class to process an input object.
        """
        pass

    def receive_complete(self):
        """Implemented by a command class to do any cleanup required
        after all input has been received.
        """
        if self._receiver:
            self._receiver.receive_complete()


    # BaseOp compile-time interface
    
    def connect(self, new_op):
        last_op  = self
        while last_op._next_op:
            last_op = last_op._next_op
        last_op._next_op = new_op
        return self


    # For use by this package

    def set_parent(self, parent):
        self._parent = parent


#--------------------------------------------------------------------------------

class Op(BaseOp):
    """Base class for all osh commands, (excluding pipelines).
    """

    # state
    
    _args = None


    # object interface
    
    def __init__(self, flag_spec, anon_range):
        BaseOp.__init__(self)
        if osh_usage == USAGE_API:
            self._args = args.ArgsAPI(self, flag_spec, anon_range)
        elif osh_usage == USAGE_CLI:
            self._args = args.ArgsCLI(self, flag_spec, anon_range)
        else:
            assert False

    def __str__(self):
        return '%s#%s%s' % (self._command_name(), self.id(), self._args)


    # BaseOp interface

    def execute(self):
        assert False


    # Op compile-time interface
    
    def add_arg(self, arg):
        self._args.add_token(arg)
        return self

    def args_done(self):
        self._args.args_done()
        return self

        
    # Op interface - for use by subclasses

    def process_args(self, *args):
        if osh_usage == USAGE_API:
            for arg in args:
                self._args.add_arg(arg)
            self._args.args_done()
            # print 'process_args: %s' % self._args
        # else: args will be supplied from parser using add_arg and
        # processed by setup method.
        return self

    def args(self):
        return self._args
    
    
    # for use by this class

    def _command_name(self):
        classname = str(self.__class__)
        # classname is <class 'MODULE.CLASSNAME'>. First get rid of everything bug MODULE.CLASSNAME.
        first_quote = classname.find("'")
        assert first_quote > 0, classname
        second_quote = classname.find("'", first_quote + 1)
        assert second_quote > 0, classname
        classname = classname[first_quote + 1 : second_quote]
        # Convert _Foo to foo
        classname = classname.split('.')[-1]
        assert len(classname) >= 2, classname
        assert classname[0] == '_', classname
        assert classname[1].isupper(), classname
        return classname[1:].lower()

#--------------------------------------------------------------------------------

class Pipeline(BaseOp):

    # state

    _first_op = None
    

    # object interface

    def __init__(self, op = None):
        BaseOp.__init__(self)
        if op:
            self.append_op(op)

    def __str__(self):
        return 'pipeline#%s' % self.id()


    # OshObject interface

    def dump(self):
        return '%s(%s)' % (self, self._first_op.dump())


    # BaseOp runtime interface

    def setup(self):
        op = self._first_op
        while op:
            op.setup()
            next = op._next_op
            if next:
                op._receiver = next
            else:
                op._receiver = self._pipeline_receiver()
            op = next

    def execute(self):
        try:
            self._first_op.execute()
        except error.ExceptionHandlerException:
            raise
        except Exception, e:
            error.exception_handler(e, self._first_op, None)

    def receive(self, object):
        self._first_op.receive(object)

    def receive_complete(self):
        self._first_op.receive_complete()
    

    # Pipeline compile-time interface

    def append_op(self, op):
        if self._first_op:
            self._first_op.connect(op)
        else:
            self._first_op = op
        op.set_parent(self)
        return self

    def prepend_op(self, op):
        if self._first_op:
            self._first_op = op.connect(self._first_op)
        else:
            self._first_op = op
        op.set_parent(self)
        return self


    # For use by this class

    def _pipeline_receiver(self):
        receiver = self._next_op
        if not receiver:
            parent = self._parent
            if parent:
                receiver = parent._pipeline_receiver()
            else:
                receiver = None
        return receiver


 #--------------------------------------------------------------------------------

class Generator(Op):
    """Base class of osh commands that are generators, e.g. gen, timer.
    """

    # object interface

    def __init__(self, flag_spec, anon_range):
        Op.__init__(self, flag_spec, anon_range)


    # BaseOp interface

    def generator(self):
        return True

    # Allows one generator to be run under the control of another
    # (e.g. timer)
    def receive(self, object):
        self.execute()


    # Generator interface

    def execute(self):
        assert False

#--------------------------------------------------------------------------------

class command(object):
    _pipeline = None

    def __init__(self, pipeline):
        self._pipeline = pipeline

    def dump(self):
        return self._pipeline.dump()

    def execute(self):
        # Prepare for execution.
        self._pipeline.setup()
        if verbosity >= 1:
            print self._pipeline.dump()
        # Execute.
        self._pipeline.execute()

    def pipeline(self):
        return self._pipeline

#--------------------------------------------------------------------------------

def _print_command_output(command, stdout, stderr):
    if stdout:
        print 'stdout from %s' % command
        for line in stdout:
            sys.stdout.write(line)
    if stderr:
        print 'stderr from %s' % command
        for line in stderr:
            sys.stdout.write(line)

class _RemoteOpThread(threading.Thread):

    _remoteop = None
    _host = None

    def __init__(self, remoteop, host):
        threading.Thread.__init__(self)
        self._remoteop = remoteop
        self._host = host

    def run(self):
        try:
            self._remoteop.action(self._host)
        except Exception, e:
            if verbosity > 0:
                traceback.print_exc()
            assert False, e

    def id(self):
        return self._id

    def host(self):
        return self._host

    def terminating_exception(self):
        return self._remoteop.terminating_exception()


class RemoteOp(Generator):
    """Base class of osh commands that run remotely. Such commands must
    be generators, (might want to permit piping of input to remote commands
    someday). Remote execution is done across all nodes in parallel.
    """

    _cluster = None
    _scp_options = None
    _terminating_exception = None

    # object interface
    
    def __init__(self, flag_spec, anon_range):
        Generator.__init__(self, flag_spec, anon_range)
        

    # BaseOp interface

    def setup(self, cluster_required = True):
        args = self.args()
        cluster_name = args.string_arg('-c')
        if cluster_name:
            # cluster_name should be one of the following:
            # - cluster defined in profile, e.g. foo
            # - cluster and pattern, e.g. foo:116
            cluster_name, pattern = (cluster_name.split(':') + [None])[:2]
            self._cluster = cluster.cluster_named(cluster_name)
            if pattern:
                self._cluster = self._cluster.filter(pattern)
        else:
            # Cluster not specified.
            if cluster_required:
                # Use default in .oshrc.
                cluster_name = config_value('remote')
                if cluster_name:
                    self._cluster = cluster.cluster_named(cluster_name)
        # Get scp options
        self._scp_options = ''
        for flag in SCP_OPTIONS:
            self._check_option(flag)
        if self._scp_options:
            self._scp_options = '-' + self._scp_options
        # Check that we have everything
        if cluster_required and not self._cluster:
            self.usage()


    # Generator interface

    def execute(self):
        threads = []
        id = 0
        for host in self.hosts():
            thread = _RemoteOpThread(self, host)
            thread.start()
            threads.append(thread)
        for thread in threads:
            while thread.isAlive():
                thread.join(1.0)
                thread_termination = thread.terminating_exception()
                if thread_termination:
                    error.exception_handler(thread_termination, self, None, thread.host().name())


    # RemoteOp interface

    def action(self):
        """Method executed on each remote node.
        """
        assert False

    def terminating_exception(self):
        return self._terminating_exception
    
    def set_terminating_exception(self, terminating_exception):
        self._terminating_exception = terminating_exception


    # For use by subclasses

    def cluster(self):
        return self._cluster

    def user(self):
        return self._cluster.user

    def hosts(self):
        return self._cluster.hosts

    def scp_options(self):
        return self._scp_options

    # For use by this class

    def _check_option(self, option):
        if self.args().flag('-%s' % option):
            self._scp_options += option

class RemoteException(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)

#
# Namespace of osh command
#

_namespace = oshrc_symbols.copy()
        
def add_to_namespace(key, value):
    _namespace[key] = value

def namespace():
    return _namespace
